#include "VoterSim.h"
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "ResultFile.h"
#include "DBResultFile.h"
#include "VotingSystem.h"

VoterSim::VoterSim()
    : numv( 100 ), numc( 4 ), trials( 10000 ), printVoters( false ), 
    printAllResults( false ), text( stdout ), doError( false ),
    confusionError( -1.0 ),
    voterDumpPrefix( NULL ), voterDumpPrefixLen( 0 ),
    voterBinDumpPrefix( NULL ), voterBinDumpPrefixLen( 0 ),
    resultDump( NULL ), resultDumpHtml( NULL ),
    summary( 1 ),
    resultDumpHtmlVertical( 0 ), winners( NULL )
{
}
VoterSim::~VoterSim() {
}

char* optstring = "c:D:d:e:F:H:h:L:l:n:qPR:rv:";


int VoterSim::init( int argc, char** argv ) {
    int c;
    int sys;
    extern char* optarg;
    
    while ( (c = getopt(argc,argv,optstring)) != -1 ) switch ( c ) {
    case 'v':	// number of voters
        numv = atoi( optarg );
        break;
    case 'n':	// number of trials
        trials = atoi( optarg );
        break;
    case 'c':	// number of candidates
        numc = atoi( optarg );
        break;
    case 'P':	// print all voters for each trial
        printVoters = ! printVoters;
        break;
    case 'r':	// print results for all trails (not just summary)
        printAllResults = ! printAllResults;
        break;
    case 'R':	// result dump
	resultDump = fopen( optarg, "w" );
	if ( resultDump == NULL ) {
	    perror( argv[0] );
	    fprintf( stderr, "opening result dump file \"%s\" failed\n", optarg );
	}
	break;
    case 'H':	// result dump html
	resultDumpHtmlVertical = 1;
    case 'h':	// result dump html
	resultDumpHtml = fopen( optarg, "w" );
	if ( resultDumpHtml == NULL ) {
	    perror( argv[0] );
	    fprintf( stderr, "opening html result dump file \"%s\" failed\n", optarg );
	}
	break;
    case 'e':	// apply error to voter preferences
        doError = true;
        confusionError = atof( optarg );
        if ( confusionError < 0.0 ) {
            confusionError *= -1.0;
        }
        if ( ! (confusionError <= 2.0) ) {
            confusionError = 2.0;
        }
        break;
    case 'D':	// dump voters, binary
	voterBinDumpPrefix = optarg;
	voterBinDumpPrefixLen = strlen(voterBinDumpPrefix);
	break;
    case 'd':	// dump voters, text
	voterDumpPrefix = optarg;
	voterDumpPrefixLen = strlen(voterDumpPrefix);
	break;
    case 'L': {	// load voters, binary
	FILE* fin = fopen( optarg, "r" );
	if ( !fin ) {
	    perror( optarg );
	    break;
	}
	fread( &numv, sizeof(numv), 1, fin );
	fread( &numc, sizeof(numc), 1, fin );
	they.build( numv, numc );
	for ( int v = 0; v < numv; v++ ) {
	    they[v].read( fin, numc );
	}
	fclose( fin );
	trials = 1;
    }	break;
    case 'l': {	// load voters, text
	FILE* fin = fopen( optarg, "r" );
	if ( !fin ) {
	    perror( optarg );
	    break;
	}
	fscanf( fin, "%d", &numv );
	fscanf( fin, "%d", &numc );
	they.build( numv, numc );
	for ( int v = 0; v < numv; v++ ) {
	    they[v].undump( fin, numc );
	}
	fclose( fin );
	trials = 1;
    }	break;
    case 'q':	// quiet, no summary
	summary = 0;
	break;
    default:
	printf("unknown option \'%c\'\n", c );
    }
    
    if ( winners != NULL ) delete [] winners;
    winners = new int[numc];

    happisum = new double[nsys];
    happistdsum = new double[nsys];
    happiness = new double*[nsys];
    for ( sys = 0; sys < nsys; sys++ ) {
        happiness[sys] = new double[trials];
    }
    return 0;
}

extern int goGently;

u_int32_t voterSteps[] = { 10, 100, 1000, 10000, 100000 };
u_int32_t choiceSteps[] = { 2, 3, 4, 5, 6, 7, 8, 9, 11, 13, 16, 20, 25, 31, 38, 50, 65, 99 };
float errorSteps[] = { -1, 0, 0.01, 0.10, 0.25, 0.5, 1.0, 1.5, 2.0 };

void VoterSim::runSteps( DBResultFile* drf, NameBlock& nb ) {
    int vstep, cstep, estep;
    for ( vstep = 0; vstep < (int)(sizeof(voterSteps)/sizeof(u_int32_t)); vstep++ ) {
	numv = voterSteps[vstep];
	for ( cstep = 0; cstep < (int)(sizeof(choiceSteps)/sizeof(u_int32_t)); cstep++ ) {
	    numc = choiceSteps[cstep];
	    they.clear();
	    if ( doError ) {
		theyWithError.clear();
	    }
    	    for ( estep = 0; estep < (int)(sizeof(errorSteps)/sizeof(float)); estep++ ) {
		confusionError = errorSteps[estep];
		run( drf, nb );
		if ( goGently ) return;
	    }
	}
    }
}

void VoterSim::run( DBResultFile* drf, NameBlock& nb ) {
    Result* r = NULL;
    if ( drf != NULL ) {
	r = drf->get( numc, numv, confusionError );
    }
    if ( r == NULL ) {
	r = newResult( nsys );
    }
    run( r );
    if ( goGently ) { fprintf(stderr,"aborted\n");return; }
    printResultHeader( stdout, numc, numv, r->trials, confusionError, nsys );
    printResult( stdout, r, nb.names, nsys );
    if ( drf ) {
	drf->put( r, numc, numv, confusionError );
    }
    free( r );
}
void VoterSim::run( Result* r ) {
    int i;
    int sys;
    if ( r == NULL ) {
	return;
    }
    for ( sys = 0; sys < nsys; sys++ ) {
        happisum[sys] = 0;
        happistdsum[sys] = 0;
    }
    if ( doError ) {
	theyWithError.build( numv, numc );
    }
    if ( resultDump != NULL ) {
	for ( sys = 0; sys < nsys; sys++ ) {
	    fprintf( resultDump, "%s:happiness\t%s:winner\t", systems[sys]->name, systems[sys]->name );
	}
	fprintf( resultDump, "\n" );
    }
    if ( resultDumpHtml != NULL ) {
	if ( resultDumpHtmlVertical ) {
	    fprintf( resultDumpHtml, "<table>\n<tr><th>Method</th><th>Happiness</th><th ALIGN=\"center\">Winner</th></tr>\n" );
	} else {
	    fprintf( resultDumpHtml, "<table>\n<tr>" );
	    for ( sys = 0; sys < nsys; sys++ ) {
		fprintf( resultDumpHtml, "<th>%s:happiness</th><th>%s:winner</th>", systems[sys]->name, systems[sys]->name );
	    }
	    fprintf( resultDumpHtml, "</tr>\n" );
	}
    }
    they.build( numv, numc );
    for ( i = 0; i < trials; i++ ) {
	if ( goGently ) return;
	for ( int v = 0; v < numv; v++ ) {
	    they[v].randomize();
	}
        if ( doError ) {
            for ( int v = 0; v < numv; v++ ) {
                theyWithError[v].setWithError( they[v], confusionError );
            }
        }
        if ( printVoters ) {
            fprintf(text,"voters = ");
            for ( int v = 0; v < numv; v++ ) {
                they[v].print(text);
                if ( doError ) {
                    theyWithError[v].print(text);
                }
                putc('\n',text);
            }
        }
	if ( voterDumpPrefix != NULL ) {
	    static char* voteDumpFilename = NULL;
	    if ( voteDumpFilename == NULL ) {
		voteDumpFilename = (char*)malloc( voterDumpPrefixLen + 30 );
	    }
	    snprintf( voteDumpFilename, voterDumpPrefixLen + 30, "%s%.10d", voterDumpPrefix, i );
	    voterDump( voteDumpFilename, they, numv, numc );
	}
	if ( voterBinDumpPrefix != NULL ) {
	    static char* voteBinDumpFilename = NULL;
	    if ( voteBinDumpFilename == NULL ) {
		voteBinDumpFilename = (char*)malloc( voterBinDumpPrefixLen + 30 );
	    }
	    snprintf( voteBinDumpFilename, voterBinDumpPrefixLen + 30, "%s%.10d", voterBinDumpPrefix, i );
	    voterBinDump( voteBinDumpFilename, they, numv, numc );
	}
	if ( resultDumpHtml != NULL ) {
	    if ( ! resultDumpHtmlVertical ) {
		fprintf( resultDumpHtml, "<tr>" );
	    }
	}
        for ( sys = 0; sys < nsys; sys++ ) {
	    double td;
	    if ( goGently ) return;
            if ( doError ) {
                systems[sys]->runElection( winners, theyWithError );
            } else {
                systems[sys]->runElection( winners, they );
            }
            happisum[sys] += happiness[sys][i] = VotingSystem::pickOneHappiness( they, numv, *winners, &td );
	    happistdsum[sys] += td;
            if ( printAllResults ) {
                fprintf(text,"%s\thappiness\t%f\twinner\t%d\n", systems[sys]->name,happiness[sys][i],*winners);
//                fprintf(text,"");
            }
	    if ( resultDump != NULL ) {
		fprintf( resultDump, "%.15g\t%d\t", happiness[sys][i], *winners );
	    }
	    if ( resultDumpHtml != NULL ) {
		if ( resultDumpHtmlVertical ) {
		    fprintf( resultDumpHtml, "<tr ALIGN=\"center\"><th>%s</th><td>%.15g</td><td>%d</td></tr>\n", systems[sys]->name, happiness[sys][i], *winners );
		} else {
		    fprintf( resultDumpHtml, "<td>%.15g</td><td>%d</td>", happiness[sys][i], *winners );
		}
	    }
        }
	if ( resultDump != NULL ) {
	    fprintf( resultDump, "\n" );
	}
	if ( resultDumpHtml != NULL && ! resultDumpHtmlVertical ) {
	    fprintf( resultDumpHtml, "</tr>\n" );
	}
	if ( i+1 < trials ) {
            for ( int v = 0; v < numv; v++ ) {
		they[v].randomize();
	    }
	}
    }
    if ( resultDumpHtml != NULL ) {
	fprintf( resultDumpHtml, "</table>\n" );
	fclose( resultDumpHtml );
    }
    unsigned int ttot = trials + r->trials;
    for ( sys = 0; sys < nsys; sys++ ) {
        double variance;
	r->systems[sys].meanHappiness = (happisum[sys] + r->systems[sys].meanHappiness * r->trials) / ttot;
	r->systems[sys].consensusStdAvg = (happistdsum[sys] + r->systems[sys].consensusStdAvg * r->trials) / ttot;
	happistdsum[sys] /= trials;
        variance = r->systems[sys].reliabilityStd * r->systems[sys].reliabilityStd * r->trials;
        for ( i = 0; i < trials; i++ ) {
            double d;
            d = happiness[sys][i] - r->systems[sys].meanHappiness;
            variance += d * d;
        }
	variance /= ttot;
	r->systems[sys].reliabilityStd = sqrt( variance );
    }
    r->trials = ttot;
}
