/* This file is placed in the public domain without warrantee by it's creator, Brian Olson. */
#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <time.h>
#include <unistd.h>
#include <math.h>
#include <string.h>
#include <signal.h>

#include "Voter.h"
#include "VoterSim.h"
#include "VotingSystem.h"

#include "OneVotePickOne.h"
#include "RankedVotePickOne.h"
#include "AcceptanceVotePickOne.h"
#include "FuzzyVotePickOne.h"
#include "InstantRunoffVotePickOne.h"
#include "Condorcet.h"
#include "ResultFile.h"
#include "DBResultFile.h"

void Voter::setWithError( const Voter& it, float error ) {
    for ( int i = 0; i < preflen; i++ ) {
        preference[i] = it.preference[i] + ((random() / ((double)LONG_MAX)) - 0.5) * 2.0 * error;
        if ( preference[i] < -1.0 ) {
            preference[i] = -1.0;
        } else if ( preference[i] > 1.0 ) {
            preference[i] = 1.0;
        }
    }
}
// "pretty" text
void Voter::print( void* fi ) const {
    FILE* f = (FILE*)fi;
    fprintf(f,"( %g", preference[0]);
    for ( int i = 1; i < preflen; i++ ) {
        fprintf(f,", %g",preference[i]);
    }
    fprintf(f," )");
}
// ugly, parse friendly text
void Voter::dump( void* fi ) const {
    FILE* f = (FILE*)fi;
    fprintf(f,"%.15g", preference[0]);
    for ( int i = 1; i < preflen; i++ ) {
        fprintf(f,"\t%.15g",preference[i]);
    }
    fprintf(f,"\n");
}
// text
void Voter::undump( void* fi, int len ) {
    FILE* f = (FILE*)fi;
    if ( len != preflen || preference == NULL ) {
	if ( preference ) {
	    delete [] preference;
	}
	preference = new float[len];
	preflen = len;
    }
    for ( int i = 0; i < len; i++ ) {
	fscanf( f, "%f",&(preference[i]));
    }
}
// binary
void Voter::write( void* fi ) const {
    FILE* f = (FILE*)fi;
    fwrite( preference, sizeof(*preference), preflen, f );
}

// binary
void Voter::read( void* fi, int len ) {
    FILE* f = (FILE*)fi;
    if ( len != preflen || preference == NULL ) {
	if ( preference ) {
	    delete [] preference;
	}
	preference = new float[len];
	preflen = len;
    }
    fread( preference, sizeof(*preference), len, f );
}

int Voter::getMax() const {
    int i;
    float m = preference[0];
    int toret = 0;
    
    for ( i = 1; i < preflen; i++ ) {
        if ( preference[i] > m ) {
            m = preference[i];
            toret = i;
        }
    }
    return toret;
}

//int Voter::defaultNumCandidates = 8;

#if 01
Voter::Voter()
    : preference( NULL ), preflen( 0 ) {
};
#else
Voter::Voter()
    : preflen( defaultNumCandidates ) {
    preference = new float[preflen];
};
#endif

Voter::Voter( int numCandidates )
    : preflen( numCandidates ) {
    preference = new float[preflen];
}
// -1.0 .. 1.0
double* prefBias;
inline double biasedRandom( double bias ) {
    // [-1.0 .. 1.0] ==> [Pi/2 .. 0]
    bias = (1.0 - bias) * (M_PI_4);
    // random in [0.0 .. 1.0] biased by exponentiation to same range, expanded to [-1.0 .. 1.0]
    return (pow( (random() / ((double)LONG_MAX)), tan( bias ) ) - 0.5) * 2.0;
}

void Voter::randomize() {
    for ( int i = 0; i < preflen; i++ ) {
#if 01
        preference[i] = ((random() / ((double)LONG_MAX)) - 0.5) * 2.0;
#elif 0
        
#else
        // scale [0.0 - 1.0) to [-1.0 - 1.0)
        preference[i] = (drand48() - 0.5) * 2.0;
#endif
    }
}

void Voter::randomizeWithFirstChoice( int first ) {
    int i;
    for ( i = 0; i < preflen; i++ ) {
#if 01
        preference[i] = ((random() / ((double)LONG_MAX)) - 0.5) * 2.0;
#elif 0
        
#else
        // scale [0.0 - 1.0) to [-1.0 - 1.0)
        preference[i] = (drand48() - 0.5) * 2.0;
#endif
    }
    // sort highest value into position of first choice
    for ( i = 0; i < preflen; i++ ) if ( i != first ) {
	if ( preference[i] > preference[first] ) {
	    float t = preference[i];
	    preference[i] = preference[first];
	    preference[first] = t;
	}
    }
}

Voter::~Voter() {
    if ( preference ) {
	delete preference;
    }
}

void Voter::setStorage( float* p, int pl ) {
    preference = p;
    preflen = pl;
}

void Voter::clearStorage() {
    preference = NULL;
    preflen = 0;
}


/* for float p[l]; return index of max */
#if 0
int getMax( float* p, int l ) {
    int i;
    float m = p[0];
    int toret = 0;
    
    for ( i = 1; i < l; i++ ) {
        if ( p[i] > m ) {
            m = p[i];
            toret = i;
        }
    }
    return toret;
}
#endif

/* v in -1.0 to 1.0 */
double quantize( double v, int steps ) {
    steps--;
    v = v + 1.0;
    v = v / 2.0;
    v = v * steps;
#ifdef __linux__
    v = rint(v);
#else
    v = round(v);
#endif
    v = v / ((double)steps);
    v = v * 2.0;
    v = v - 1.0;
    return v;
}

double VotingSystem::pickOneHappiness( const VoterArray& they, int numv, int winner, double* stddevP ) {
    double happiness = 0;
    int i;
    for ( i = 0; i < numv; i++ ) {
        happiness += they[i].getPref( winner );
    }
    happiness = happiness / numv;
    if ( stddevP != NULL ) {
	double stddev = 0;
	for ( i = 0; i < numv; i++ ) {
	    double d;
	    d = they[i].getPref( winner ) - happiness;
	    stddev += ( d * d );
	}
	stddev = stddev / numv;
	stddev = sqrt( stddev );
	*stddevP = stddev;
    }
    return happiness;
}

void voterDump( char* voteDumpFilename, const VoterArray& they, int numv, int numc ) {
    FILE* voteDumpFile;
    fprintf( voteDumpFile, "%d\n%d\n", numv, numc );
    voteDumpFile = fopen( voteDumpFilename, "w" );
    for ( int v = 0; v < numv; v++ ) {
	they[v].dump( voteDumpFile );
    }
    fclose( voteDumpFile );
}
void voterBinDump( char* voteBinDumpFilename, const VoterArray& they, int numv, int numc ) {
    FILE* voteDumpFile;
    voteDumpFile = fopen( voteBinDumpFilename, "w" );
    fwrite( &numv, sizeof(numv), 1, voteDumpFile );
    fwrite( &(numc), sizeof(numc), 1, voteDumpFile );
    for ( int v = 0; v < numv; v++ ) {
	they[v].write( voteDumpFile );
    }
    fclose( voteDumpFile );
}

// inline function to avoid double evaluation of a or b
inline double fmax( double a, double b ) {
    return (a>b)?a:b;
}

class MaxHappiness : public VotingSystem {
public:
MaxHappiness() : VotingSystem("Max Happiness") {};
virtual void runElection( int* winnerR, const VoterArray& they ) {
    double maxh;
    int c = 0;
    int numc = they.numc;
    int numv = they.numv;
    
    maxh = pickOneHappiness( they, numv, c );
    for ( int i = 1; i < numc; i++ ) {
        double h;
        h = pickOneHappiness( they, numv, i );
        if ( h > maxh ) {
            c = i;
            maxh = h;
        }
    }
    if ( winnerR ) *winnerR = c;
};
};

int goGently = 0;

void mysigint( int a ) {
    goGently = 1;
}

int main( int argc, char** argv ) {

VotingSystem* systems[] = {
    new MaxHappiness(),
    new OneVotePickOne(),
    new InstantRunoffVotePickOne(),
    new AcceptanceVotePickOne(),
    new FuzzyVotePickOne( 0, 0 ),
    new FuzzyVotePickOne( 1, 0 ),
    new FuzzyVotePickOne( 2, 0 ),
    new FuzzyVotePickOne( 0, -1 ),
    new FuzzyVotePickOne( 0, 10 ),
    new RankedVotePickOne(),
    new RankedVotePickOne("Ranked Vote, no neg pref",1),
    new Condorcet(),
    NULL
};
    VoterSim s;
    DBResultFile* drf = NULL;
    NameBlock nb;
    int j = 0;
    int nsys;

    srandom(time(NULL));

    nsys = 0;
    {VotingSystem** curSys = systems; while ( *curSys ) {
        curSys++;
        nsys++;
    }}
    votingSystemArrayToNameBlock( &nb, systems, nsys );

    for ( int i = 0; i < argc; i++ ) {
	if ( j != i ) {
	    argv[j] = argv[i];
	}
	if ( ! strcmp( argv[i], "-F" ) ) {
	    i++;
	    drf = DBResultFile::open( argv[i] );
	    i++;
	    if ( drf ) {
		drf->useNames( &nb );
	    }
	} else {
	    j++;
	}
    }
    argc = j;
    
    signal( SIGINT, mysigint );
    
    s.systems = systems;
    s.nsys = nsys;
    
    s.init( argc, argv );
    s.run( drf, nb );
    if ( drf ) drf->close();
    return 0;
}
